/** \file keyfile.c	Functions to manipulate sets of key/value pairs. Provides
 * two main objects: The key file (kf_xxx finctions), and the key value 
 * (kv_xxx functions). Only the key file functions are used outside this file,
 * but the key value functions could be exported if required. 
 *
 * GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2003 Darryl Luff
 * $Id: keyfile.c,v 1.3 2003/01/30 14:26:21 djluff Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "keyfile.h"

/** The KeyVal structure and kv_xxx routines implement a linked list.	*/
typedef struct _keyval {
	char *key;	/**< Key string.	*/
	char *val;	/**< Value string.	*/
	void *next;	/**< Pointer to the next KeyVal structure.	*/
} KeyVal;

KeyVal *kv_add(KeyVal * kv, char *k, char *v);
void kv_free(KeyVal * kv);
char *kv_getval(KeyVal * kv, char *k);
KeyVal *kv_new(char *k, char *v);
void kv_setval(KeyVal * kv, char *k, char *v);

/* Key file functions	*/
char *kf_getvalue(void *kf, char *k);
void kf_load(void *kf, char *fn);
void *kf_new(char sep);
void kf_save(void *kf, char *fn);
void kf_setvalue(void *kf, char *k, char *v);
char *trim(char *s);

/* Implementation	*/
/** Add a new KeyVal to the end of the list. If 'kv' is NULL, a new list is
 * created. Returns the root of the list (either kv or the new root).	*/
KeyVal *kv_add(KeyVal * kv, char *k, char *v)
{
	/* If no kv passed, just allocate a new one */
	if (!kv)
		return kv_new(k, v);

	/* If the passed kv is not end of the list, pass on the request to it */
	if (kv->next)
		return kv_add(kv->next, k, v);

	/* Otherwise, add a new one to the passed KeyVal */
	kv->next = kv_new(k, v);
	return kv;
}

/** Frees the KeyVal structure and all following KeyVals in the list	*/
void kv_free(KeyVal * kv)
{
	if (!kv)
		return;

	if (kv->next)
		kv_free(kv->next);

	if (kv) {
		if (kv->key)
			free(kv->key);
		if (kv->val)
			free(kv->val);
		free(kv);
	}
}

/** Searches the list rooted at kv, and returns the value for the first 
 * matching key	*/
char *kv_getval(KeyVal * kv, char *k)
{
	if (!kv)
		return NULL;

	if (strcmp(kv->key, k) == 0)
		return kv->val;
	else if (kv->next)
		return kv_getval(kv->next, k);

	return NULL;
}


/** Allocate a new KeyVal.	*/
KeyVal *kv_new(char *k, char *v)
{
	KeyVal *kv = malloc(sizeof(KeyVal));
	if (kv) {
		kv->key = k;
		kv->val = v;
		kv->next = NULL;
	}
	return kv;
}

/** Sets the passed value. This call is propogated along thr list.	*/
void kv_setval(KeyVal * kv, char *k, char *v)
{
	if (!kv)
		return;

	if (strcmp(kv->key, k) == 0) {
		if (kv->val)
			free(kv->val);
		kv->val = strdup(v);
		return;
	}

	if (kv->next)
		kv_setval(kv->next, k, v);
	else							   /* end of list and hasnt been found */
		kv_add(kv, k, v);
}

/** Write the keyVal list to a file.	*/
void kv_write(KeyVal * kv, char sep, FILE * f)
{
	if (!kv)
		return;

	if (kv->key) {
		if (kv->val)
			fprintf(f, "%s%c%s\n", kv->key, sep, kv->val);
		else
			fprintf(f, "%s\n", kv->key);
	}
	else
		fprintf(f, "\n");

	if (kv->next)
		kv_write(kv->next, sep, f);
}

/** A key file structure.	*/
typedef struct _keyfile {
	int changed;				/**< Count of changes since the file was saved.	*/
	char separator;				/**< Character that separates each line into key | value.	*/
	KeyVal *keys;				/**< The key/value list.	*/
} KeyFile;

/** Returns the value of the passed variable.	*/
char *kf_getvalue(void *kf, char *k)
{
	KeyFile *kff = (KeyFile *) kf;
	if (!kff)
		return NULL;

	return kv_getval(kff->keys, k);
}

/** Loads a keyfiles contents from disk.	*/
void kf_load(void *kf, char *fn)
{
	char *p;
	FILE *f;
	char *s;
	char buf[2048];
	KeyFile *kff = (KeyFile *) kf;
	if (!kff)
		return;

	f = fopen(fn, "r+");
	while (fgets(buf, sizeof(buf), f)) {
		p = strchr(buf, kff->separator);
		if (p) {
			*p = '\0';
			s = trim(buf);
			*p = kff->separator;
			p++;
			p = trim(p);
			kff->keys = kv_add(kff->keys, s, p);
			free(p);
			free(s);
		}
		else {						   /* separator not found */
			kff->keys = kv_add(kff->keys, buf, NULL);
		}
	}
	fclose(f);
	kff->changed = 0;
}

/** Allocate a new key file.	*/
void *kf_new(char sep)
{
	KeyFile *kf = malloc(sizeof(KeyFile));
	if (kf) {
		kf->changed = 0;
		kf->separator = sep;
		kf->keys = NULL;
	}
	return kf;
}

/** Save the keylist to the file.	*/
void kf_save(void *kf, char *fn)
{
	FILE *f;
	KeyFile *kff = (KeyFile *) kf;
	if (!kff)
		return;

	f = fopen(fn, "w");
	kv_write(kff->keys, kff->separator, f);
	fclose(f);
	kff->changed = 0;
}

/** Sets the value of the variable. This call is propagated along the keylist.	*/
void kf_setvalue(void *kf, char *k, char *v)
{
	KeyFile *kff = (KeyFile *) kf;
	if (!kff)
		return;

	kv_setval(kff->keys, k, v);
}

/** Trims whitespace from the start and end of a string.	*/
char *trim(char *s)
{
	char *end;
	char *start = s;

	while (isspace(*start) && (*start != '\0')) {
		start++;
	}
	end = start;
	/* find end */
	while (*end != '\0')
		end++;

	/* now work back, dropping space chars */
	while ((isspace(*end)) && (end >= start))
		*end = '\0';

	return strdup(start);
}
